# vim: tabstop=4 shiftwidth=4 softtabstop=4
# Copyright 2016 Cumulus Networks, Inc. All rights reserved.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc.
# 51 Franklin Street, Fifth Floor
# Boston, MA  02110-1301, USA.
"""This is the code to populate the redis DB. This code specifically assumes
that the backend DB is redis since it extensively makes use of the redis data
structures. For the scaling purposes desired by this tool, no other KV store
seemed appropriate and so the code is hardcoded to assume redis.
"""
from __future__ import absolute_import

import argparse
import bisect
import hashlib
import json
import os
import re
import socket
import subprocess
import sys
import time

import eventlet
import eventlet.pools
from eventlet.green.subprocess import TimeoutExpired
import greenlet

from netq.common import netlink, utils
from netq.common.enums import (
    Daemons,
    DefaultPaths,
    NetDevFmt,
    Heartbeat,
    NeighborState
)
from netq.common.service import Service
from netq.common.utils import NetlinkCopy
from netq.orm import HOSTNAME
from netq.orm.redisdb import (
    curr_pipeline,
    new_pipeline,
    reset_connection_settings,
    scripts,
    set_connection_settings,
    RedisException)
from netq.orm.redisdb.models import (
    Address,
    Command,
    Counter,
    Neighbor,
    Link,
    VirtualDevLink,
    Node,
    MacFdb,
    Route,
    VxlanLink
)

_PROC_NAME = 'netq-agent'


class _Command(object):
    """Command class.
    """
    # pylint: disable=too-few-public-methods
    def __init__(self, period, key, command):
        self.period = period
        self.key = key
        self.command = command

    def __cmp__(self, other):
        return cmp(self.period, other.period)

    def __eq__(self, other):
        return self.key == other.key and self.command == other.command

    def __ne__(self, other):
        return not self == other

    def __repr__(self):
        return '%s(period=%d, key=%s, command=%s)' % (
            self.__class__.__name__, self.period, self.key, self.command
        )


class _Thread(object):
    """Thread class.
    """
    def __init__(self):
        self.__function = None
        self.__pool = None
        self.__start_time = None
        self.__stop_checker = None
        self.green_thread = None
        self.pipeline = None

    def kill(self):
        eventlet.kill(self.green_thread)

    def link(self, stop_checker):
        self.__stop_checker = stop_checker
        self.green_thread.link(stop_checker)

    @classmethod
    def spawn(cls, epoch, pool, function, pipeline=None):
        thread = cls()
        thread.__function = function
        thread.__pool = pool
        thread.__start_time = epoch
        thread.pipeline = pipeline
        if pipeline is not None:
            thread.green_thread = pool.spawn(function, pipeline)
        else:
            thread.green_thread = pool.spawn(function)
        return thread

    def stale(self, now):
        return now - self.__start_time >= Heartbeat.HEARTBEAT_DEATH

    def respawn(self, epoch):
        self.kill()
        thread = self.spawn(epoch, self.__pool, self.__function,
                            pipeline=self.pipeline)
        return thread


class _ClDbPush(Service):
    """
    This class defines the main workhorse class of this file.
    It parses the args, fires off the command event monitoring
    threads, and processes the outputs posted by these threads
    and adds them to a redis database.
    """
    __BRIDGE_FBD_REGEX = (
        r'(?:(?P<deleted>^Deleted)\s+)?'
        r'(?P<mac_address>\S+)'
        r'\s+\S+\s+(?P<nexthop>\S+)'
        r'(?=.*vlan\s+(?P<vlan_id>\d+))?'
        r'(?=.*\s+(?P<origin>permanent$))?'
        r'(?=.*dst\s+(?P<dst>\S+))?'
        r'(?=.*master\s+(?P<port>\S+))?'
    )

    __COUNTERS_RATE = 30

    __CMD_TIMEOUT = 5

    __GC_RATE = 5 * 60

    __MAX_REDIS_CONNECTIONS = 10

    __PIPELINE_RATE = 2

    __REDISTRIBUTE_NBR_TABLE_ID = 10

    def __init__(self, conf):
        super(_ClDbPush, self).__init__(conf, _PROC_NAME)
        self.__nlmonitor = netlink.Netlink(
            self._conf,
            self._logger,
            {netlink.Netlink.NEWADDR: self.__rx_rtm_newaddr,
             netlink.Netlink.DELADDR: self.__rx_rtm_deladdr,
             netlink.Netlink.NEWLINK: self.__rx_rtm_newlink,
             netlink.Netlink.DELLINK: self.__rx_rtm_dellink,
             netlink.Netlink.NEWNEIGH: self.__rx_rtm_newneigh,
             netlink.Netlink.DELNEIGH: self.__rx_rtm_delneigh,
             netlink.Netlink.NEWROUTE: self.__rx_rtm_newroute,
             netlink.Netlink.DELROUTE: self.__rx_rtm_delroute},
            pool=self._pool
        )
        self.__commands = []
        self.__gc_keys = None
        self.__gc_ready = True
        self.__heartbeat_gt = None
        self.__hostname = HOSTNAME
        self.__ifname_by_index = None
        self.__lastboot = time.time()
        self.__neighbor_cache = None
        self.__prev_cmd = {}
        self.__redispool = eventlet.pools.TokenPool(
            max_size=self.__MAX_REDIS_CONNECTIONS
        )
        self.__redisthreads = set()
        self.__reinitialization = False
        self.__timers = {}

    def __bridge_monitor(self, cmd_output):
        """Processes the output from 'bridge monitor fdb'.
        The formats to consider are:
        [00:00:00:00:00:00 dev vx-14567100 dst 192.168.1.1 self permanent]
        [Deleted 00:00:00:00:00:00 dev vx-14567100 dst 192.168.1.1 self
         permanent]
        :param cmd_output:
        :return:
        """
        brmatch = re.match(self.__BRIDGE_FBD_REGEX, cmd_output)
        if brmatch is not None:
            brgrpdict = brmatch.groupdict()
            deleted = brgrpdict['deleted'] is not None
            bridge = MacFdb()
            bridge.mac_address = brgrpdict['mac_address']
            bridge.nexthop = brgrpdict['nexthop']
            bridge.origin = brgrpdict['origin'] is not None
            if brgrpdict['dst'] is not None:
                bridge.dst = str(brgrpdict['dst'])
            if brgrpdict['port'] is not None:
                bridge.port = str(brgrpdict['port'])
            bridge.vlan = int(brgrpdict['vlan_id'] or 0)
            if deleted:
                bridge.delete()
            else:
                bridge.save()

    def __gc_collect(self):
        if self.__gc_ready:
            # Make sure that the gc_ready flag is set to False before calling
            # this method
            return
        args = [self._conf.entry_retention_count]
        if self.__gc_keys is None:
            self.__gc_keys = []
            cmd = Command()
            for command in self.__commands:
                cmd.key = command.key
                self.__gc_keys.append(Command.obj_to_key(cmd.__dict__))
        if self.__gc_keys:
            try:
                freed = scripts.gc_script(self.__gc_keys, args)
                if freed > 0:
                    self._logger.info('gc_collect freed %s entries', freed)
            except RedisException:
                pass
        self.__gc_ready = True

    def __get_daemons(self):
        cmd = 'sudo vtysh -c "show daemons"'
        try:
            first_line = next(
                (line for line in
                 subprocess.check_output(cmd, shell=True).splitlines()),
                None
            )
            if first_line:
                return first_line.split()
        except subprocess.CalledProcessError as ex:
            self._logger.error('Failed to execute command %s. out %s',
                               cmd, ex.output)
        return []

    def __ifupdown_managed_interface(self, ifname):
        """Return True if this interface is managed by ifupdown2
        ifquery is slow, it takes ~250ms
        subprocess.check_output(['/sbin/ifquery', ifname])

        grep takes 2ms... we only care if the string is present or not
        so use -q so grep will exit as soon as it finds it
        """
        retcode = None
        command = '/bin/grep -q auto %s$ /etc/network/interfaces' % ifname
        cmd = subprocess.Popen(command,
                               stdout=subprocess.PIPE,
                               stderr=subprocess.STDOUT,
                               shell=True)
        try:
            # Instead of a blocking OS call, this version of wait() uses logic
            # borrowed from the eventlet 0.2 processes.Process.wait() method.
            endtime = time.time() + self.__CMD_TIMEOUT
            while True:
                status = cmd.poll()
                if status is not None:
                    retcode = status
                    break
                if time.time() > endtime:
                    raise TimeoutExpired(command, self.__CMD_TIMEOUT)
                eventlet.sleep(1)
        except TimeoutExpired:
            self._logger.error('Failed to execute command %s after %d seconds',
                               command, self.__CMD_TIMEOUT)
        except Exception:  # pylint: disable=broad-except
            self._logger.exception('Failed to execute command %s', command)
        else:
            if retcode != 0:
                self._logger.debug('Command %s returned %d', command, retcode)
            return retcode == 0
        return False

    def __load_commands(self):
        """Determine what info to push based on box configuration
        """
        daemons = [Daemons.CLAG, Daemons.MISC, Daemons.LLDP]
        if os.path.isfile(self._conf.dbconfig):
            with open(self._conf.dbconfig, 'r') as dbfiled:
                base_config_json = json.load(dbfiled)
                daemons += self.__get_daemons()
                for section, commands in base_config_json.iteritems():
                    if section in daemons:
                        for command in commands:
                            if all(k in command
                                   for k in ('period', 'key', 'command')):
                                cmd = _Command(int(command['period']),
                                               str(command['key'].strip()),
                                               str(command['command'].strip()))
                                idx = bisect.bisect_left(self.__commands, cmd)
                                if (idx >= len(self.__commands) or
                                        self.__commands[idx] != cmd):
                                    bisect.insort(self.__commands, cmd)
                            else:
                                self._logger.info('Ignoring invalid line in '
                                                  'netq-agent-commands.json:'
                                                  '%r', command)

        # The purpose of the DBCONFIG_RUNNING file is to share state between
        # netq-agent and netq.
        with open(DefaultPaths.DBCONFIG_RUNNING, 'w') as dbfiled:
            json.dump({'server': self._conf.server,
                       'commands': self.__commands},
                      dbfiled,
                      cls=utils.ClassEncoder)

        epoch = int(time.time())
        for cmd in self.__commands:
            self.__timers.setdefault(epoch + cmd.period, set())
            self.__timers[epoch + cmd.period].add(cmd)

    def __load_fdb_table(self):
        cmd = '/sbin/bridge fdb show'
        try:
            bridgecmd = subprocess.Popen(cmd.split(),
                                         stdout=subprocess.PIPE)
        except Exception:  # pylint: disable=broad-except
            self._logger.exception('Failed to read bridge fdb table')
            return
        for line in iter(bridgecmd.stdout.readline, b''):
            self.__bridge_monitor(line)

    def __pipe_execute(self, pipeline):
        with self.__redispool.item() as _:
            self._logger.debug('Executing pipeline %s of length %d',
                               id(pipeline), len(pipeline.command_stack))
            pipeline.execute()

    def __process_command(self, info):
        """This updates a key based on new data from executing the command
        :param info: Command object
        :type info: _Command
        """
        # Execute the command:
        cmd_out = retcode = None
        cmd = subprocess.Popen(info.command,
                               stdout=subprocess.PIPE,
                               stderr=subprocess.STDOUT,
                               shell=True)
        try:
            # Instead of a blocking OS call, this version of wait() uses logic
            # borrowed from the eventlet 0.2 processes.Process.wait() method.
            endtime = time.time() + self.__CMD_TIMEOUT
            while True:
                status = cmd.poll()
                if status is not None:
                    retcode = status
                    break
                if time.time() > endtime:
                    raise TimeoutExpired(info.command, self.__CMD_TIMEOUT)
                eventlet.sleep(1)
        except TimeoutExpired:
            self._logger.error('Failed to execute command %s after %d seconds',
                               info.command, self.__CMD_TIMEOUT)
        except Exception:  # pylint: disable=broad-except
            self._logger.exception('Failed to execute command %s',
                                   info.command)
        else:
            cmd_out = cmd.stdout.read()
            if retcode != 0:
                self._logger.debug('Command %s returned %d', info.command,
                                   retcode)
        # Don't push the data out unless it has changed
        if cmd_out:
            cmd = Command()
            cmdhash = hashlib.md5(cmd_out).digest()
            prev_cmd = self.__prev_cmd.get(info.key, None)
            if prev_cmd is None or prev_cmd[0] != cmdhash:
                self.__prev_cmd[info.key] = (cmdhash, len(cmd_out))
                cmd.key = info.key
                cmd.output = cmd_out
                cmd.save()
        epoch = int(time.time())
        self.__timers.setdefault(epoch + info.period, set())
        self.__timers[epoch + info.period].add(info)

    def __process_counters(self):
        """Update Top talkers and such
        Push out counters per port so they can be sorted by redis.
        """
        cmd = 'cat /proc/net/dev | tail -n +3'
        netstat_cmd = subprocess.Popen(cmd, shell=True,
                                       stdout=subprocess.PIPE,
                                       stderr=subprocess.STDOUT)
        output, _ = netstat_cmd.communicate()
        for line in output.splitlines():
            if not line.strip():
                continue
            words = line.split()
            counter = Counter()
            counter.ifname = words[NetDevFmt.IFACE].split(':')[0]
            counter.tx_bytes = int(words[NetDevFmt.TX_BYTES])
            counter.rx_bytes = int(words[NetDevFmt.RX_BYTES])
            counter.tx_drop_bytes = int(words[NetDevFmt.TX_DRP])
            counter.rx_drop_bytes = int(words[NetDevFmt.RX_DRP])
            counter.save()

    def __reinitialize_db(self):
        if self.__reinitialization:
            self._logger.info('Reinitialization underway')
            return
        self._logger.info('Reinitializing the database')
        self.__reinitialization = True
        # Flush the pipeline
        self._logger.debug('Flushing the pipeline')
        for entry in list(self.__redisthreads):
            entry.kill()
        self.__redisthreads = set()
        curr_pipeline().reset()

        # Delete all the netlink keys
        self._logger.debug('Deleting netlink keys')
        timeout = 1
        while True:
            try:
                Address.query.delete(hostname=HOSTNAME, purge=False)
                Link.query.delete(hostname=HOSTNAME, purge=False)
                Neighbor.query.delete(hostname=HOSTNAME, purge=False)
                MacFdb.query.delete(hostname=HOSTNAME, purge=False)
                Route.query.delete(hostname=HOSTNAME, purge=False)
                VxlanLink.query.delete(hostname=HOSTNAME, purge=False)
                break
            except RedisException:
                reset_connection_settings()
                timeout *= 2
                self._logger.exception('Delete failed, retrying after %d '
                                       'seconds', timeout)
                eventlet.sleep(timeout)

        # Send a heartbeat
        self._logger.debug('Sending heartbeat to server')
        self.__update_nodename()

        # Reinitialize the bridge fdb table
        self._logger.debug('Loading bridge fdb table')
        self.__load_fdb_table()

        # Rebind to netlink
        self._logger.debug('Binding to netlink')
        if self.__nlmonitor.socket is not None:
            self._logger.info('Rebinding netlink socket')
        timeout = 1
        while True:
            try:
                self.__nlmonitor.bind()
                break
            except RuntimeError:
                timeout *= 2
                self._logger.info('Rebinding netlink socket after %d seconds',
                                  timeout)
                eventlet.sleep(timeout)
        self._pool.spawn_n(self._serve,
                           self.__nlmonitor.socket,
                           self.__nlmonitor.handle_netlink_msg,
                           bufsize=netlink.Netlink.NLSOCK_BYTES,
                           err_cbs={
                               socket.errno.ENOBUFS: self.__reinitialize_db
                           })
        self.__ifname_by_index = {}
        self.__neighbor_cache = set()
        self._pool.spawn(self.__nlmonitor.get_all_links).wait()
        self._pool.spawn_n(self.__nlmonitor.get_all_addresses)
        self._pool.spawn_n(self.__nlmonitor.get_all_neighbors)
        self._pool.spawn_n(self.__nlmonitor.get_all_routes)
        self.__reinitialization = False

    def __rx_rtm_newaddr(self, info):
        """Adds a new address to the DB.
        :param info: address object
        :type info: netlink.Address
        """
        self._logger.info('RX RTM_NEWADDR for ip_address %s, ifindex %s',
                          '/'.join((info.prefix, str(info.mask))),
                          info.ifindex)
        addr = Address()
        NetlinkCopy.copy_address(addr, info, self.__ifname_by_index)
        self._logger.debug('Adding new address %r to DB', addr)
        addr.save()
        return info.ifindex

    def __rx_rtm_deladdr(self, info):
        """Deletes an address from the DB.
        :param info: address object
        :type info: netlink.Address
        """
        self._logger.info('RX RTM_DELADDR for ip_address %s, ifindex %s',
                          '/'.join((info.prefix, str(info.mask))),
                          info.ifindex)
        addr = Address()
        NetlinkCopy.copy_address(addr, info, self.__ifname_by_index)
        self._logger.debug('Deleting address %r from DB', addr)
        addr.delete()
        return info.ifindex

    def __rx_rtm_newlink(self, info):
        """Adds a new link to the DB.
        :param info: link object
        :type info: netlink.Link
        """
        self._logger.info('RX RTM_NEWLINK for ifname %s, admin_state %s, '
                          'oper_state %s', info.ifname, info.admin_state,
                          info.oper_state)
        self.__ifname_by_index[info.ifindex] = info.ifname
        if isinstance(info, netlink.VxlanLink):
            link = VxlanLink()
        elif info.kind == 'bond' or info.kind == 'bridge':
            link = VirtualDevLink()
        else:
            link = Link()
        NetlinkCopy.copy_link(link, info)
        link.managed = self.__ifupdown_managed_interface(link.ifname)
        self._logger.debug('Adding new link %r to DB', link)
        link.save()
        return info.ifindex

    def __rx_rtm_dellink(self, info):
        """Deletes a link from the DB.
        :param info: link object
        :type info: netlink.Link
        """
        self._logger.info('RX RTM_DELLINK for ifname %s, admin_state %s, '
                          'oper_state %s', info.ifname, info.admin_state,
                          info.oper_state)
        if info.kind in ['bond', 'bridge']:
            VirtualDevLink.query.delete(ifname=info.ifname,
                                        hostname=self.__hostname)
        elif isinstance(info, netlink.VxlanLink):
            VxlanLink.query.delete(ifname=info.ifname,
                                   hostname=self.__hostname,
                                   vni=info.vni)
        else:
            Link.query.delete(ifname=info.ifname,
                              kind=info.kind,
                              hostname=self.__hostname)
        return info.ifindex

    def __rx_rtm_newneigh(self, info):
        """Adds a new neighbor to the DB.
        Neighbors in the kernel go through a lot of state transitions that
        we are currently not interested in tracking.  Only update redis if a
        neighbor is new...ie ignore all of the RTM_NEWNEIGH messages
        generated from the state transitions
        :param info: neighbor object
        :type info: netlink.Neighbor
        """
        self._logger.info('RX RTM_NEWNEIGH for ip_address %s, ifindex %s',
                          info.ip_address, info.ifindex)
        neighbor = Neighbor()
        NetlinkCopy.copy_neighbor(neighbor, info, self.__ifname_by_index)
        key = hashlib.md5(Neighbor.obj_to_key(neighbor.__dict__)).digest()
        if (info.state == NeighborState.UPDATE and
                key not in self.__neighbor_cache):
            self._logger.debug('Adding new neighbor %r to DB', neighbor)
            self.__neighbor_cache.add(key)
            neighbor.save()
        elif info.state == NeighborState.DELETE:
            self._logger.debug('Deleting neighbor %r from DB', neighbor)
            neighbor.delete()
        return info.ifindex

    def __rx_rtm_delneigh(self, info):
        """Deletes a neighbor from the DB.
        :param info: neighbor object
        :type info: netlink.Neighbor
        """
        self._logger.info('RX RTM_DELNEIGH for ip_address %s, ifindex %s',
                          info.ip_address, info.ifindex)
        neighbor = Neighbor()
        NetlinkCopy.copy_neighbor(neighbor, info, self.__ifname_by_index)
        self._logger.debug('Deleting neighbor %r from DB', neighbor)
        neighbor.delete()
        return info.ifindex

    def __rx_rtm_newroute(self, info):
        """Adds a new route to the DB.
        :param info: route object
        :type info: netlink.Route
        """
        # Skip redistribute neighbor routes in table 10
        if info.table == self.__REDISTRIBUTE_NBR_TABLE_ID:
            return info.ifindex

        self._logger.info('RX RTM_NEWROUTE for ip_address %s, nexthops %s',
                          info.ip_address, info.nexthops)
        route = Route()
        NetlinkCopy.copy_route(route, info, self.__ifname_by_index)
        self._logger.debug('Adding new route %r to DB', route)
        route.save()
        return info.ifindex

    def __rx_rtm_delroute(self, info):
        """Delete a route from the DB.
        :param info: route object
        :type info: netlink.Route
        """
        self._logger.info('RX RTM_DELROUTE for ip_address %s, nexthops %s',
                          info.ip_address, info.nexthops)
        route = Route()
        NetlinkCopy.copy_route(route, info, self.__ifname_by_index)
        self._logger.debug('Deleting route %r from DB', route)
        route.delete()
        return info.ifindex

    def __stop_thread_checker(self, green_thread):
        """ Propagates exceptions to the run green thread.
        """
        entry = next((thread for thread in self.__redisthreads
                      if thread.green_thread == green_thread), None)
        if entry is not None:
            self.__redisthreads.remove(entry)
            if entry == self.__heartbeat_gt:
                self.__heartbeat_gt = None
        try:
            green_thread.wait()
        except greenlet.GreenletExit:  # pylint: disable=no-member
            pass
        except RedisException:
            if entry is not None:
                if entry.pipeline is not None:
                    self._logger.exception('Failed to execute PIPELINE %s',
                                           id(entry.pipeline))
                    self._pool.spawn_n(self.__reinitialize_db)
                else:
                    self._logger.exception('Failed to send heartbeat')
        except Exception:  # pylint: disable=broad-except
            eventlet.kill(self._run_gt, *sys.exc_info())
        else:
            if entry is not None:
                if entry.pipeline is not None:
                    self._logger.debug('Successfully executed PIPELINE %s',
                                       id(entry.pipeline))
                else:
                    self._logger.debug('Successfully sent heartbeat')

    def __update_nodename(self):
        node = Node()
        node.lastboot = self.__lastboot
        with self.__redispool.item() as _:
            self._logger.info('Sending heartbeat to server')
            node.save(commit=True)

    def _run(self):
        """Once upon a time...
        """
        set_connection_settings(self._conf.server, agent=True)
        self._logger.info('%s started, Redis server %s', _PROC_NAME,
                          self._conf.server)

        # Purge and reinit entries
        self.__reinitialize_db()
        # Start the netlink dispatcher
        self.__nlmonitor.run().link(self._stop_checker)
        # Assemble the data to be pushed
        self.__load_commands()

        # Queue up the initial set of MAC entries.
        self._pool.spawn_n(self._serve_command,
                           '/sbin/bridge --oneline monitor fdb',
                           self.__bridge_monitor)
        timers = {'pipeline': 0, 'counters': 0, 'gc': 0, 'liveliness': 0,
                  'inactivity': 0}
        while True:
            epoch = int(time.time())
            new_timers, self.__timers = self.__timers.copy(), {}
            for key in new_timers:
                if epoch > key:
                    for cmd in new_timers[key]:
                        green_thread = self._pool.spawn(self.__process_command,
                                                        cmd)
                        green_thread.link(self._stop_checker)
                else:
                    self.__timers[key] = new_timers[key]
            if epoch - timers['inactivity'] >= Heartbeat.HEARTBEAT_DEATH:
                for thread in list(self.__redisthreads):
                    if thread.stale(epoch):
                        if thread == self.__heartbeat_gt:
                            self._logger.warn('Killing heartbeat thread')
                            self.__heartbeat_gt.kill()
                        else:
                            self._logger.warn('Inactivity timeout exceeded '
                                              'for pipeline %s. Respawning...',
                                              id(thread.pipeline))
                            new_thread = thread.respawn(epoch)
                            new_thread.link(self.__stop_thread_checker)
                            self.__redisthreads.add(new_thread)
                timers['inactivity'] = epoch
            if epoch - timers['liveliness'] >= Heartbeat.LIVELINESS_RATE:
                if self.__heartbeat_gt is None:
                    self.__heartbeat_gt = _Thread.spawn(epoch,
                                                        self._pool,
                                                        self.__update_nodename)
                    self.__heartbeat_gt.link(self.__stop_thread_checker)
                    self.__redisthreads.add(self.__heartbeat_gt)
                timers['liveliness'] = epoch
            if epoch - timers['pipeline'] >= self.__PIPELINE_RATE:
                pipeline = curr_pipeline()
                if pipeline.command_stack:
                    new_pipeline()
                    thread = _Thread.spawn(epoch,
                                           self._pool,
                                           self.__pipe_execute,
                                           pipeline=pipeline)
                    thread.link(self.__stop_thread_checker)
                    self.__redisthreads.add(thread)
                timers['pipeline'] = epoch
            if epoch - timers['counters'] >= self.__COUNTERS_RATE:
                self.__process_counters()
                timers['counters'] = epoch
            if epoch - timers['gc'] >= self.__GC_RATE and self.__gc_ready:
                self.__gc_ready = False
                green_thread = self._pool.spawn(self.__gc_collect)
                green_thread.link(self._stop_checker)
                timers['gc'] = epoch
            eventlet.sleep(1)


def main():
    # command line args
    parser = argparse.ArgumentParser(
        description='%s: daemon that implements pushes interesting '
                    'data to a distributed KV store' % _PROC_NAME,
        epilog='This is experimental code',
    )
    parser.add_argument('-d', '--daemon', action='store_true')
    parser.add_argument('-c', '--config-file', type=str,
                        default=DefaultPaths.CONF_FILE)
    parser.add_argument('-p', '--pidfile', type=str,
                        default=DefaultPaths.PID_FILE)
    parser.add_argument('-s', '--server', default=None, type=str)
    parser.add_argument('-D', '--debug', action='store_true')
    parser.add_argument('-v', '--version', action='version',
                        version='%(prog)s 0.91')
    args = parser.parse_args()
    conf = utils.load_configuration(args)
    cldbpush_inst = _ClDbPush(conf)
    return cldbpush_inst.run()
