# vim: tabstop=4 shiftwidth=4 softtabstop=4
# Copyright 2016 Cumulus Networks, Inc. All rights reserved.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc.
# 51 Franklin Street, Fifth Floor
# Boston, MA  02110-1301, USA.
from __future__ import absolute_import

import calendar
import json
import re
import socket
import subprocess
import sys
import time
from datetime import datetime, timedelta
from itertools import chain

import ipaddr
from tabulate import tabulate

from netq.common import config, utils
from netq.common.enums import (
    BgColors,
    Heartbeat,
    DefaultPaths
)
from netq.lib.parsers import (
    build_bgp_summary,
    build_clag_map,
    build_lldp_summary
)
from netq.orm import HOSTNAME
from netq.orm import redisdb, restapi
from netq.orm.memdb.models import (
    Lldp,
    MemNode,
    BgpSession
)
from netq.orm.query import Q
from netq.orm.redisdb.models import (
    Address,
    Command,
    Counter,
    Link,
    VirtualDevLink,
    MacFdb,
    Neighbor,
    Node,
    Route,
    VxlanLink
)
from netq.orm.redisdb.query import RedisQuery


def get_hostname():
    # deal with -s <server-ip>|<hostname> view [around <time>] <host> <cmd>
    if sys.argv[1] == 'server':
        offset = 2
    else:
        offset = 0

    if sys.argv[offset + 1] == 'view':
        if sys.argv[offset + 2] == 'around':
            hostname = sys.argv[offset + 4]
        else:
            hostname = sys.argv[offset + 2]
    else:
        # all other commands have the hostname just before the
        # current arg
        hostname = sys.argv[-2]
    return hostname


class NetQException(Exception):
    def __init__(self, *args, **kwargs):
        super(NetQException, self).__init__(*args, **kwargs)


class NetQ(object):
    __INTF_REGEX = re.compile(
        r'^\d+: (?P<dev_name>\S+):\s+'
        r'<.*(?P<slave>SLAVE).*>'
        r'(?=.*inet (?P<ipv4>\S+) scope (?P<scopev4>\S+))'
        r'(?=.*inet6 (?P<ipv6>\S+) scope (?P<scopev6>\S+))'
        r'(?=.*master (?P<master>\S+))'
        r'(?=.*link/ether (?P<mac>\S+))'
    )

    __IP_ADDR_REGEX = re.compile(
        r'\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}(?:/\d+)?'
    )

    __TIME_REGEX = re.compile(
        r'(?:(?:(?P<hours>[\d|\.]+)h)|'
        r'(?:(?P<minutes>[\d|\.]+)m)|'
        r'(?:(?P<seconds>[\d|\.]+)s))'
    )

    __TRACEROUTE_REGEX = r'\s*(?P<idx>\d+).*'

    ADDR_KEY = 'addr'
    BGP_SUMMARY_KEY = 'bgp-summary-json'
    BGP_NEIGHBOR_KEY = 'bgp-neighbors'
    CLAGCTL_JSON_KEY = 'clagctl-json'
    LLDP_NEIGHBOR_KEY = 'lldp-neighbor'
    UPTIME_KEY = 'uptime'
    MEMINFO_KEY = 'meminfo'

    def __init__(self, server_addr, start_time):
        self.__node_lst = None
        self.failed_node_lst = set()

        self.__server_addr = (
            self.__get_server_address(server_addr) or
            self.__get_config_server_addr()
        )

        self.__start_time = self.__get_start_time(start_time)
        if server_addr is not None and 'http' in server_addr:
            restapi.set_connection_settings(self.__server_addr)
        else:
            redisdb.set_connection_settings(self.__server_addr)

    def __build_bgp_summ(self, bad_sessions=False):
        """Build summary of BGP node info, such as ASN & RID
        """
        for node in self.node_lst:
            last_beat = (datetime.utcnow() -
                         datetime.utcfromtimestamp(node.timestamp))
            if last_beat.total_seconds() > Heartbeat.HEARTBEAT_DEATH:
                self.failed_node_lst.add(node.hostname)
                continue
            elif node.hostname in self.failed_node_lst:
                self.failed_node_lst.remove(node.hostname)

            cmd = Command.query.get(timestamp=self.__start_time,
                                    key=NetQ.BGP_SUMMARY_KEY,
                                    hostname=node.hostname)
            if cmd is not None:
                neigh_out = None
                if bad_sessions:
                    neigh_cmd = Command.query.get(timestamp=self.__start_time,
                                                  key=NetQ.BGP_NEIGHBOR_KEY,
                                                  hostname=node.hostname)
                    if neigh_cmd is not None:
                        neigh_out = neigh_cmd.output
                try:
                    build_bgp_summary(node.hostname,
                                      cmd.timestamp,
                                      cmd.output,
                                      neigh_output=neigh_out)
                except Exception:  # pylint: disable=broad-except
                    continue

    def __build_clag_map(self):
        """
        """
        for node in self.node_lst:
            last_beat = (datetime.utcnow() -
                         datetime.utcfromtimestamp(node.timestamp))
            if last_beat.total_seconds() > Heartbeat.HEARTBEAT_DEATH:
                self.failed_node_lst.add(node.hostname)
                continue
            elif node.hostname in self.failed_node_lst:
                self.failed_node_lst.remove(node.hostname)

            cmd = Command.query.get(timestamp=self.__start_time,
                                    hostname=node.hostname,
                                    key=NetQ.CLAGCTL_JSON_KEY)
            if cmd is not None:
                try:
                    build_clag_map(node.hostname, cmd.timestamp, cmd.output)
                except Exception:  # pylint: disable=broad-except
                    continue

    def __build_lldp_info(self, hostname=None):
        if hostname is not None:
            nodes = sorted(Node.query.filter(hostname=hostname))
        else:
            nodes = self.node_lst
        for node in nodes:
            last_beat = (datetime.utcnow() -
                         datetime.utcfromtimestamp(node.timestamp))
            if last_beat.total_seconds() > Heartbeat.HEARTBEAT_DEATH:
                self.failed_node_lst.add(node.hostname)
                continue
            elif node.hostname in self.failed_node_lst:
                self.failed_node_lst.remove(node.hostname)

            cmd = Command.query.get(timestamp=self.__start_time,
                                    key=NetQ.LLDP_NEIGHBOR_KEY,
                                    hostname=node.hostname)
            if cmd is not None:
                try:
                    build_lldp_summary(node.hostname, cmd.timestamp,
                                       cmd.output)
                except Exception:  # pylint: disable=broad-except
                    continue

    @staticmethod
    def __get_config_server_addr():
        try:
            conf = config.Config(DefaultPaths.CONF_FILE)
            if conf.server:
                return conf.server
        except Exception:  # pylint: disable=broad-except
            pass
        try:
            with open(DefaultPaths.DBCONFIG_RUNNING, 'r') as fd:
                base_config_json = json.load(fd)
                if 'server' in base_config_json:
                    return (
                        str(ipaddr.IPv4Network(base_config_json['server']).ip)
                    )
        except Exception:  # pylint: disable=broad-except
            raise RuntimeError('Please specify IP address of DB server')

    @staticmethod
    def __get_server_address(server_addr):
        try:
            return str(ipaddr.IPAddress(server_addr))
        except Exception:  # pylint: disable=broad-except
            try:
                return socket.gethostbyname(server_addr)
            except Exception:  # pylint: disable=broad-except
                return server_addr

    def __get_start_time(self, start_time):
        if start_time is None:
            return start_time
        match = self.__TIME_REGEX.match(start_time)
        if match is not None:
            groupdict = match.groupdict()
            return calendar.timegm(
                (
                    datetime.utcnow() -
                    timedelta(
                        hours=float(groupdict['hours'] or 0),
                        minutes=float(groupdict['minutes'] or 0),
                        seconds=float(groupdict['seconds'] or 0)
                    )
                ).utctimetuple()
            )
        try:
            return calendar.timegm(
                datetime.strptime(start_time,
                                  '%Y-%m-%dT%H:%M:%S.%fZ').utctimetuple()
            )
        except Exception:  # pylint: disable=broad-except
            try:
                return calendar.timegm(
                    datetime.strptime(start_time, '%Y-%m-%d').utctimetuple()
                )
            except Exception:  # pylint: disable=broad-except
                try:
                    datetime.utcfromtimestamp(start_time)
                    return start_time
                except Exception:  # pylint: disable=broad-except
                    return None

    def path_trace_mac(self, mac, vlan, start):
        """
        """
        output = []
        for entry in MacFdb.query.filter(timestamp=self.__start_time,
                                         hostname=start,
                                         mac_address=mac,
                                         vlan=vlan):
            output.append(entry.hostname)
            """
            FIXME: Dinesh
            for oif in self.macvlanmap[key][start]:
                if oif in self.__linkmap[start].keys():
                    if 'slaves' in self.__linkmap[start][oif].keys():
                        for slave in self.__linkmap[start][oif]['slaves']:
                            if slave in self.nbrmap[start].keys():
                                nbrnode = \
                                    self.nbrmap[start][slave]['nbr-node']
                                iface = "%s-%s" % (oif, slave)
                                output.append([iface,
                                               self.__path_trace_mac(mac,
                                                                     vlan,
                                                                     nbrnode)])
                            else:
                                output.append([oif, ''])
                    else:
                        if oif in self.nbrmap[start].keys():
                            nbrnode = self.nbrmap[start][oif]['nbr-node']
                            output.append([oif,
                                         self.__path_trace_mac(mac,
                                                               vlan,
                                                               nbrnode)])
                        else:
                            output.append([oif, ''])
        if output is not None:
            utils.prtree(output, 0, True)
            print
        """

    def resolve_output(self):
        """Replace the IP address in output with node name & interface
        if possible.
        """
        prev = HOSTNAME
        traceroute = False
        pmtu_failure = False
        mtu_string = ''
        out = []
        for line in sys.stdin:
            addresses = {}
            pmtu_error = ''
            prev_save = None
            if not traceroute and line.startswith('traceroute'):
                self.__build_lldp_info()
                traceroute = True
                mtu = 0
                mtu_string = ''

            for addr in self.__IP_ADDR_REGEX.findall(line):
                try:
                    if '/' in addr:
                        _ = ipaddr.IPNetwork(addr)
                        prefix, mask = addr.split('/')
                        addrobj = Address.query.get(
                            timestamp=self.__start_time,
                            prefix=prefix,
                            mask=mask
                        )
                    else:
                        _ = ipaddr.IPAddress(addr)
                        addrobj = Address.query.get(
                            timestamp=self.__start_time,
                            prefix=addr
                        )
                except Exception:  # pylint: disable=broad-except
                    continue

                if addrobj is None:
                    continue

                if traceroute and prev is not None:
                    pmtu_error = ''                    
                    lldp = Lldp.query.get(hostname=prev,
                                          lldp_peer=addrobj.hostname)
                    if lldp is not None:
                        local_link = Link.query.get(
                            timestamp=self.__start_time,
                            ifname=lldp.local_intf,
                            hostname=prev)
                        if local_link:
                            if mtu == 0:
                                mtu = int(local_link.mtu)
                                mtu_string = '%s:%s, MTU %d' % (prev,
                                                                lldp.local_intf,
                                                                mtu)

                            if mtu != local_link.mtu:
                                pmtu_error = '*'
                                pmtu_failure = True
                        remote_link = Link.query.get(
                            timestamp=self.__start_time,
                            ifname=lldp.lldp_intf,
                            hostname=lldp.lldp_peer
                        )
                        if remote_link and mtu != int(remote_link.mtu):
                            pmtu_error = '*'
                            pmtu_failure = True

                        if '*' in pmtu_error:
                            addr_tuple = (
                                utils.color_wrap(BgColors.RED,
                                                 lldp.lldp_peer),
                                utils.color_wrap(BgColors.RED,
                                                 lldp.lldp_intf),
                                pmtu_error)
                        else:
                            addr_tuple = (
                                utils.color_wrap(BgColors.GREEN,
                                                 lldp.lldp_peer),
                                utils.color_wrap(BgColors.GREEN,
                                                 lldp.lldp_intf))

                        addresses[addr] = addr_tuple
                    else:
                        addr_tuple = (
                            utils.color_wrap(BgColors.GREEN, addrobj.hostname),
                            utils.color_wrap(BgColors.GREEN, addrobj.ifname)
                        )
                        addresses[addr] = addr_tuple
                else:
                    addr_tuple = (
                        utils.color_wrap(BgColors.GREEN, addrobj.hostname),
                        utils.color_wrap(BgColors.GREEN, addrobj.ifname)
                    )
                    addresses[addr] = addr_tuple

                if not line.startswith('traceroute') and not prev_save:
                    prev_save = addrobj.hostname

            for addr in addresses:
                addr_str = '%s (%s)' % (addr, ':'.join(addresses[addr]))
                if traceroute:
                    line = line.replace(addr, addr_str, 1)
                    line = line.replace(' (%s)' % addr, '', 1)
                else:
                    line = line.replace(addr, addr_str)
            out.append(line)

            if not line.startswith('traceroute'):
                prev = prev_save

        if pmtu_failure:
            out.append(
                '* - Links with mismatched MTU from origin %s' %
                mtu_string
            )
        elif traceroute:
            if mtu == 0:
                out.append('Path MTU couldn\'t be determined')
            else:
                out.append('Path MTU is %d' % mtu)

        return ''.join(out)

    def show_address_history(self, node, iface, ip_address, mac, ipv6=None):
        out = []
        if not ipv6:
            if ip_address == '*':
                ipv6 = '*'
            else:
                ipv6 = ':' in ip_address

        now = time.time()
        if '/' in ip_address:
            prefix, mask = ip_address.split('/')
            active_entries = Address.query.rangebyscore(
                prefix=prefix, mask=mask, active=True, is_ipv6=ipv6,
                hostname=node, ifname=iface,
                start=self.__start_time or 0, end=now, first=False
            )
            inactive_entries = Address.query.rangebyscore(
                prefix=prefix, mask=mask, active=False, is_ipv6=ipv6,
                hostname=node, ifname=iface,
                start=self.__start_time or 0, end=now, first=False
            )
        else:
            active_entries = Address.query.rangebyscore(
                prefix=ip_address, active=True, is_ipv6=ipv6,
                hostname=node, ifname=iface,
                start=self.__start_time or 0, end=now, first=False
            )
            inactive_entries = Address.query.rangebyscore(
                prefix=ip_address, active=False, is_ipv6=ipv6,
                hostname=node, ifname=iface,
                start=self.__start_time or 0, end=now, first=False
            )
        for address in sorted(list(active_entries) + list(inactive_entries),
                              key=lambda x: x.timestamp,
                              reverse=True):
            out.append(('/'.join((address.prefix, str(address.mask))),
                        address.hostname, address.ifname, address.active,
                        utils.pretty_date(address.timestamp)))
        return out

    def show_addresses(self, node, iface, ip_address, mac):
        want_ipv6 = not ip_address or ':' in ip_address
        if '/' in ip_address:
            prefix, mask = ip_address.split('/')
        else:
            prefix = ip_address
            mask = '*'

        # Ripping out MAC-based query support for now as it wasn't working
        if ip_address == '*':
            entries = Address.query.filter(timestamp=self.__start_time,
                                           hostname=node, ifname=iface,
                                           prefix=prefix, mask=mask)
        else:
            entries = Address.query.filter(timestamp=self.__start_time,
                                           hostname=node, ifname=iface,
                                           prefix=prefix, mask=mask,
                                           is_ipv6=want_ipv6)
        out = []
        for address in entries:
            ipa = address.prefix + '/' + str(address.mask)
            loopback_addr = ['127.0.0.1/8', '::1/128']
            if ipa in loopback_addr:
                continue
            out.append(('/'.join((address.prefix, str(address.mask))),
                        address.hostname, address.ifname,
                        utils.pretty_date(address.timestamp)))
            if len(out) == RedisQuery.PAGE_SIZE:
                yield out
                out = []
        if out:
            yield out

    @staticmethod
    def show_agent_history(hostname=None):
        if hostname is not None:
            query = Node.query.range(start=0,
                                     end=-1,
                                     first=False,
                                     reverse=True,
                                     hostname=hostname)
        else:
            query = Node.query.range(start=0,
                                     end=-1,
                                     first=False,
                                     reverse=True)
        for obj in query:
            yield (obj.hostname,
                   datetime.utcfromtimestamp(int(obj.lastboot)),
                   datetime.utcfromtimestamp(int(obj.timestamp)))

    def show_asn(self, asn):
        self.__build_bgp_summ()

        out = []
        if asn is None:
            nodes = MemNode.query.all(timestamp=self.__start_time)
        else:
            nodes = MemNode.query.filter(timestamp=self.__start_time,
                                         bgp_asn=asn)
        for node in nodes:
            out.append((node.bgp_asn, node.hostname, node.bgp_router_id))
        return out

    def show_bgp_status(self):
        out = []
        self.__build_bgp_summ(bad_sessions=True)
        min_timestamp = time.time()

        for session in BgpSession.query.filter(timestamp=self.__start_time,
                                               exclude=Q(reason='')):
            out.append([utils.color_wrap(BgColors.RED, session.hostname),
                        utils.color_wrap(BgColors.RED, session.peer_name),
                        utils.color_wrap(BgColors.RED, session.reason)])

            if min_timestamp == 0.0 or session.timestamp < min_timestamp:
                min_timestamp = session.timestamp

        if out != []:
            out.append(
                utils.color_wrap(BgColors.BLUE, 'Output retrieved around %s' %
                                 (utils.pretty_date(min_timestamp)))           
            )

        return out

    def show_clag_status(self):
        self.__build_clag_map()
        failed_node_lst = []

        for clag in MemNode.query.filter(timestamp=self.__start_time,
                                         clag_state=False):
            if not clag.clag_peer_state:
                failed_node_lst.append([utils.color_wrap(BgColors.RED,
                                                         clag.hostname),
                                        'Peer Connectivity failed'])

            if clag.clag_conflicted_bonds != []:
                err_str = 'Conflicted Bonds: %s' % (clag.clag_conflicted_bonds)
                failed_node_lst.append([utils.color_wrap(BgColors.RED,
                                                         clag.hostname),
                                        err_str])
            if clag.clag_proto_down_bonds != []:
                err_str = 'ProtoDown Bonds: %s' % (clag.clag_proto_down_bonds)
                failed_node_lst.append([utils.color_wrap(BgColors.RED,
                                                         clag.hostname),
                                        err_str])

        for node in self.failed_node_lst:
            failed_node_lst.append([utils.color_wrap(BgColors.RED,
                                                     node),
                                    'Rotten Agent State'])
        return failed_node_lst

    def show_info(self, ip_address=None, mac_address=None, node=None):
        """
        Get all relevant info about provided input
        """
        display = []
        if ip_address is not '*':
            self.__build_lldp_info(None)
            self.__build_bgp_summ(False)
            # Attempt to determine if we can identify the end node
            out = []
            neighbors = Neighbor.query.filter(timestamp=self.__start_time,
                                              ip_address=ip_address)
            for neighbor in neighbors:
                macs = MacFdb.query.filter(timestamp=self.__start_time,
                                           mac_address=neighbor.mac_address)
                for entry in macs:
                    lldpnbrs = Lldp.query.filter(timestamp=self.__start_time,
                                                 hostname=neighbor.hostname,
                                                 local_intf=entry.nexthop)
                    for lldpnbr in lldpnbrs:
                        display.append(
                            '%s belongs to %s, connected to %s:%s\n' % (
                                ip_address, lldpnbr.lldp_peer,
                                neighbor.hostname, entry.nexthop)
                        )
                out.append((neighbor.ip_address, neighbor.hostname,
                            neighbor.ifname, neighbor.mac_address))
                if out:
                    display.append('IP Neighbor Information:')
                    display.append(tabulate(sorted(out),
                                            headers=['IP', 'Node', 'Interface',
                                                     'MAC']))

        elif mac_address is not '*':
            mac_entries = MacFdb.query.filter(timestamp=self.__start_time,
                                              mac_address=mac_address)
            for entry in mac_entries:
                lldpnbrs = Lldp.query.filter(timestamp=self.__start_time,
                                             hostname=entry.hostname,
                                             local_intf=entry.nexthop)
                for lldpnbr in lldpnbrs:
                    display.append('%s belongs to %s, connected to %s:%s\n' %
                                   (mac_address, lldpnbr.lldp_peer,
                                    entry.hostname, entry.nexthop))
            neighbors = Neighbor.query.filter(timestamp=self.__start_time,
                                              mac_address=mac_address,
                                              is_ipv6=False)
            out = []
            for neighbor in neighbors:
                out.append((neighbor.ip_address, neighbor.hostname,
                            neighbor.ifname, neighbor.mac_address))
            if out:
                display.append('List of IP Neighbor entries')
                display.append(tabulate(sorted(out),
                                        headers=['IP', 'Node', 'Interface',
                                                 'MAC']))

        elif node is not '*':
            if Node.query.get(hostname=node) is None:
                raise NetQException('No information about %s found' % node)

            # OK, this is a hostname. Display all valid info about host
            display.append('Info about node %s' % node)
            services = []
            cmd = Command.query.get(timestamp=self.__start_time,
                                    hostname=node,
                                    key='bgp-summary*')
            if cmd and 'bgpd is not running' not in cmd.output:
                services.append('BGP')
                self.__build_bgp_summ(node)

            cmd = Command.query.get(timestamp=self.__start_time,
                                    hostname=node,
                                    key='clagctl*')
            if cmd and 'Unable to communicate ' not in cmd.output:
                services.append('CLAG')
                self.__build_clag_map()

            cmd = Command.query.get(timestamp=self.__start_time,
                                    hostname=node,
                                    key='ospf*')
            if cmd and 'ospfd is not running' not in cmd.output:
                services.append('OSPF')

            cmd = Command.query.get(timestamp=self.__start_time,
                                    hostname=node,
                                    key='mstp*')
            if cmd and 'Usage' not in cmd.output:
                services.append('MSTP')

            cmd = Command.query.get(timestamp=self.__start_time,
                                    hostname=node,
                                    key='lldp-neighbor')
            if cmd and 'unable to connect' not in cmd.output:
                services.append('LLDP')

            uptime = Command.query.get(timestamp=self.__start_time,
                                       hostname=node,
                                       key=NetQ.UPTIME_KEY)
            if uptime is not None:
                display.append('uptime %s' % uptime.output.strip())

            meminfo = Command.query.get(timestamp=self.__start_time,
                                        hostname=node,
                                        key=NetQ.MEMINFO_KEY)
            if meminfo is not None:
                display.append('Free Memory: %s' %
                               meminfo.output.split('\n')[1].
                               split(':')[1].strip())

            routes = Route.query.filter(timestamp=self.__start_time,
                                        hostname=node)
            macs = MacFdb.query.filter(timestamp=self.__start_time,
                                       hostname=node)

            display.append('Total routes: %d' % len(list(routes)))
            display.append('Total macs: %d' % len(list(macs)))
            display.append('Services running: %s' % ', '.join(services))

            # Make sure this is the last thing probed about the node
            # We return if no node-specific info is found
            memnode = MemNode.query.get(timestamp=self.__start_time,
                                        hostname=node)
            if memnode is None:
                display.append('Warning: No node info available for services')
            else:
                if 'BGP' in services:
                    display.append('BGP: Inactive Peer Count: %d' % (
                        memnode.bgp_peer_count - memnode.bgp_active_peer_count
                    ))
                if 'CLAG' in services:
                    display.append('CLAG: Role: %s, peer status: %s' % (
                        memnode.clag_role, memnode.clag_peer_state))
        return '\n'.join(display)

    def show_ip_neighbor_history(self, hostname='*', ifname='*',
                                 ip_address='*', mac='*', ipv6=False):
        out = []
        now = time.time()
        if hostname != '*' and ifname != '*':
            active_entries = Neighbor.query.rangebyscore(
                hostname=hostname, ifname=ifname, active=True,
                mac_address=mac, is_ipv6=ipv6,
                start=self.__start_time or 0, end=now, first=False
            )
            inactive_entries = Neighbor.query.rangebyscore(
                hostname=hostname, ifname=ifname, active=False,
                mac_address=mac, is_ipv6=ipv6,
                start=self.__start_time or 0, end=now, first=False
            )
        else:
            active_entries = Neighbor.query.rangebyscore(
                ip_address=ip_address, active=True, mac_address=mac,
                is_ipv6=ipv6, start=self.__start_time or 0, end=now,
                first=False
            )
            inactive_entries = Neighbor.query.rangebyscore(
                ip_address=ip_address, active=False, mac_address=mac,
                is_ipv6=ipv6, start=self.__start_time or 0, end=now,
                first=False
            )
        for neighbor in sorted(list(active_entries) + list(inactive_entries),
                               key=lambda x: x.timestamp,
                               reverse=True):
            out.append((neighbor.ip_address, neighbor.hostname,
                        neighbor.ifname, neighbor.mac_address, neighbor.active,
                        utils.pretty_date(neighbor.timestamp)))
        return out

    def show_ip_neighbors(self, ip_address, hostname, iface, mac, ipv6=False):
        out = []
        for neighbor in Neighbor.query.filter(timestamp=self.__start_time,
                                              ip_address=ip_address,
                                              hostname=hostname,
                                              ifname=iface,
                                              mac_address=mac,
                                              is_ipv6=ipv6):
            if neighbor.ifname == 'lo':
                continue
            if neighbor.ip_address == 'None':
                continue
            out.append((neighbor.ip_address, neighbor.hostname,
                        neighbor.ifname, neighbor.mac_address,
                        utils.pretty_date(neighbor.timestamp)))
            if len(out) == RedisQuery.PAGE_SIZE:
                yield out
                out = []
        if out:
            yield out

    def show_link_history(self, hostname, ifname):
        self.__build_lldp_info(None)
        out = []
        now = time.time()

        pactive_entries = Link.query.rangebyscore(
            hostname=hostname, ifname=ifname, active=True,
            start=self.__start_time or 0, end=now, first=False
        )
        pinactive_entries = Link.query.rangebyscore(
            hostname=hostname, ifname=ifname, active=False,
            start=self.__start_time or 0, end=now, first=False
        )
        vactive_entries = VirtualDevLink.query.rangebyscore(
            hostname=hostname, ifname=ifname, active=True,
            start=self.__start_time or 0, end=now, first=False
        )
        vinactive_entries = VirtualDevLink.query.rangebyscore(
            hostname=hostname, ifname=ifname, active=False,
            start=self.__start_time or 0, end=now, first=False
        )
        vxactive_entries = VxlanLink.query.rangebyscore(
            hostname=hostname, ifname=ifname, active=True,
            start=self.__start_time or 0, end=now, first=False
        )
        vxinactive_entries = VxlanLink.query.rangebyscore(
            hostname=hostname, ifname=ifname, active=False,
            start=self.__start_time or 0, end=now, first=False
        )
        active_entries = chain(pactive_entries, vactive_entries,
                               vxactive_entries)
        inactive_entries = chain(pinactive_entries, vinactive_entries,
                                 vxinactive_entries)
        for link in sorted(list(active_entries) + list(inactive_entries),
                           key=lambda x: x.timestamp,
                           reverse=True):
            lldp_string = ''
            lldp_peer = Lldp.query.get(self.__start_time,
                                       hostname=link.hostname,
                                       local_intf=link.ifname)
            if lldp_peer is not None:
                lldp_string = '%s:%s' % (lldp_peer.lldp_peer,
                                         lldp_peer.lldp_intf)
                if link.kind == '':
                    if lldp_peer.lldp_peer_bridge:
                        link.kind = 'L2Intf'
                    elif lldp_peer.lldp_peer_router:
                        link.kind = 'L3Intf'
            out.append((link.hostname, link.ifname, link.admin_state,
                        link.oper_state, link.kind, lldp_string, link.mtu,
                        link.active, utils.pretty_date(link.timestamp)))
        return out

    def show_links(self, hostname=None, ifname=None):
        self.__build_lldp_info(None)
        plinks = Link.query.filter(timestamp=self.__start_time,
                                   hostname=hostname, ifname=ifname)
        vlinks = VirtualDevLink.query.filter(timestamp=self.__start_time,
                                             hostname=hostname, ifname=ifname)
        vxlinks = VxlanLink.query.filter(timestamp=self.__start_time,
                                         hostname=hostname, ifname=ifname)
        links = chain(plinks, vlinks, vxlinks)

        out = []
        for link in links:
            lldp_string = ''
            lldp_peer = Lldp.query.get(self.__start_time,
                                       hostname=link.hostname,
                                       local_intf=link.ifname)

            if lldp_peer is not None:
                lldp_string = '%s:%s' % (lldp_peer.lldp_peer,
                                         lldp_peer.lldp_intf)
                if link.kind == '':
                    if lldp_peer.lldp_peer_bridge:
                        link.kind = 'L2Intf'
                    elif lldp_peer.lldp_peer_router:
                        link.kind = 'L3Intf'
            out.append((link.hostname, link.ifname, link.admin_state,
                        link.oper_state, link.kind, lldp_string, link.mtu,
                        link.active, utils.pretty_date(link.timestamp)))
            if len(out) == RedisQuery.PAGE_SIZE:
                yield out
                out = []
        if out:
            yield out

    def show_mac_history(self, mac_address, vlan='*', hostname='*',
                         nexthop='*'):
        out = []
        now = time.time()
        active_entries = MacFdb.query.rangebyscore(
            mac_address=mac_address, hostname=hostname, vlan=vlan,
            nexthop=nexthop, active=True, start=self.__start_time or 0,
            end=now, first=False
        )
        inactive_entries = MacFdb.query.rangebyscore(
            mac_address=mac_address, hostname=hostname, vlan=vlan,
            nexthop=nexthop, active=False,
            start=self.__start_time or 0, end=now, first=False
        )
        for macfdb in sorted(list(active_entries) + list(inactive_entries),
                             key=lambda x: x.timestamp,
                             reverse=True):
            out.append([macfdb.mac_address, macfdb.vlan, macfdb.hostname,
                        macfdb.nexthop, macfdb.active,
                        utils.pretty_date(macfdb.timestamp)])
        return out

    def show_macvlan(self, mac_address, vlan, hostname, nexthop):
        fdb_entries = MacFdb.query.filter(timestamp=self.__start_time,
                                          hostname=hostname,
                                          mac_address=mac_address, vlan=vlan,
                                          nexthop=nexthop)
        out = []
        for macfdb in fdb_entries:
            out.append([macfdb.mac_address, macfdb.vlan, macfdb.hostname,
                        macfdb.nexthop, utils.pretty_date(macfdb.timestamp)])
            if len(out) == RedisQuery.PAGE_SIZE:
                yield out
                out = []
        if out:
            yield out

    def show_neighbors(self, node, iface):
        self.__build_lldp_info(node)
        out = []
        if iface is not '*':
            lldplist = Lldp.query.filter(timestamp=self.__start_time,
                                         hostname=node,
                                         local_intf=iface)
        else:
            lldplist = Lldp.query.all(timestamp=self.__start_time)
        for lldpobj in lldplist:
            out.append((lldpobj.hostname, lldpobj.local_intf,
                        lldpobj.lldp_peer, lldpobj.lldp_intf,
                        utils.pretty_date(lldpobj.timestamp)))
        return out

    def show_route_history(self, ip_address, hostname='*', ipv6=False):
        out = []
        now = time.time()
        active_entries = Route.query.rangebyscore(
            ip_address=ip_address, hostname=hostname, active=True,
            is_ipv6=ipv6,
            start=self.__start_time or 0, end=now, first=False
        )
        inactive_entries = Route.query.rangebyscore(
            ip_address=ip_address, hostname=hostname, active=False,
            is_ipv6=ipv6,
            start=self.__start_time or 0, end=now, first=False
        )
        for route in sorted(list(active_entries) + list(inactive_entries),
                            key=lambda x: x.timestamp,
                            reverse=True):
            nexthops = ', '.join(
                '%s: %s' % (nh[0], nh[1]) if isinstance(nh, list) else nh
                for nh in route.nexthops
            )
            out.append(
                (route.ip_address, route.hostname, nexthops,
                 route.origin, route.active,
                 utils.pretty_date(route.timestamp)))
        return out

    def show_router_id(self, router_id):
        self.__build_bgp_summ()
        out = []
        if router_id is None:
            nodes = MemNode.query.all(timestamp=self.__start_time)
        else:
            nodes = MemNode.query.filter(timestamp=self.__start_time,
                                         bgp_router_id=router_id)
        for node in nodes:
            out.append((node.bgp_router_id, node.hostname, node.bgp_asn))
        return out

    def show_routes(self, prefix, hostname, ipv6=False, origin='*'):
        skip_ipv4_addr = {'127.0.0.0/8', '127.0.0.1/32', '127.0.0.0/32',
                          '127.255.255.255/32', '169.254.0.1/32'}
        skip_ipv6_addr = {'fe80::/64', '::/0', '::1/128', 'ff00::/8'}
        if hostname is not '*':
            nodes = [hostname]
        else:
            nodes = [node.hostname for node in self.node_lst]
        if prefix is not '*':
            entries = Route.query.lpm(prefix, nodes,
                                      timestamp=self.__start_time,
                                      is_ipv6=ipv6)
        elif hostname is not '*':
            entries = Route.query.filter(timestamp=self.__start_time,
                                         hostname=hostname,
                                         is_ipv6=ipv6)
        else:
            entries = Route.query.filter(timestamp=self.__start_time,
                                         is_ipv6=ipv6)
        out = []
        for route in entries:
            if not route.is_ipv6 and route.ip_address in skip_ipv4_addr:
                continue
            if route.is_ipv6 and route.ip_address in skip_ipv6_addr:
                continue
            if not route.nexthops:
                route.nexthops = ['Local']
            if origin is '*' or origin == route.origin:
                nexthops = ', '.join(
                    '%s: %s' % (nh[0], nh[1]) if isinstance(nh, list)
                    else nh for nh in route.nexthops
                )
                out.append((route.origin,
                            route.table,
                            route.ip_address,
                            route.hostname,
                            nexthops,
                            utils.pretty_date(route.timestamp)))
                if len(out) == RedisQuery.PAGE_SIZE:
                    yield out
                    out = []
        if out:
            yield out

    def show_status(self):
        out = []
        for node in self.node_lst:
            ago = utils.pretty_date(node.timestamp)
            lastboot = datetime.utcfromtimestamp(int(node.lastboot))
            last_beat = (datetime.utcnow() -
                         datetime.utcfromtimestamp(int(node.timestamp)))
            if last_beat.total_seconds() > Heartbeat.HEARTBEAT_DEATH:
                status = utils.color_wrap(BgColors.RED, 'Rotten')
            elif last_beat.total_seconds() <= Heartbeat.LIVELINESS_RATE:
                status = utils.color_wrap(BgColors.GREEN, 'Fresh')
            else:
                status = utils.color_wrap(BgColors.YELLOW, 'Stale')
            out.append([str(node), str(lastboot), ago, status])
        return out

    @staticmethod
    def top_talkers_rx(count):
        output = []
        drop_nodes = []
        for obj in Counter.query.range(start=0,
                                       end=count,
                                       reverse=True,
                                       key_type=Counter.RX_BYTES,
                                       first=False):
            output.append((obj.hostname, obj.ifname, obj.rx_bytes))
        for obj in Counter.query.range(start=-10000,
                                       end=1,
                                       reverse=True,
                                       key_type=Counter.RX_DROP_BYTES,
                                       first=False):
            drop_nodes.append((obj.hostname, obj.ifname, obj.rx_drop_bytes))
        return output, drop_nodes

    @staticmethod
    def top_talkers_tx(count):
        output = []
        drop_nodes = []
        for obj in Counter.query.range(start=0,
                                       end=count,
                                       reverse=True,
                                       key_type=Counter.TX_BYTES,
                                       first=False):
            output.append((obj.hostname, obj.ifname, obj.tx_bytes))
        for obj in Counter.query.range(start=-10000,
                                       end=1,
                                       reverse=True,
                                       key_type=Counter.TX_DROP_BYTES,
                                       first=False):
            drop_nodes.append((obj.hostname, obj.ifname, obj.tx_drop_bytes))
        return output, drop_nodes

    def view(self, node, key):
        cmd = Command.query.get(timestamp=self.__start_time,
                                hostname=node,
                                key='-'.join(key))
        if cmd is None:
            raise NetQException(
                'Could\'nt locate command output for key %s' % ' '.join(key)
            )
        return '\n'.join([
            utils.color_wrap(BgColors.RED, 'Output retrieved from %s' %
                             utils.pretty_date(cmd.timestamp)),
            cmd.output
        ])

    def vxlan_trace(self, dst):
        """ Perform a vxlan traceroute from the destination IP address to the
        source.
        :param dst: dst ip address
        """
        cmd = '/usr/sbin/traceroute -n %s' % dst
        tracecmd1 = subprocess.Popen(cmd.split(),
                                     stdout=subprocess.PIPE,
                                     stderr=subprocess.STDOUT)
        first_line = last_line = ''
        for line in iter(tracecmd1.stdout.readline, b''):
            first_line = first_line or line
            last_line = line
        if tracecmd1.returncode == 0:
            raise NetQException('Traceroute to %s failed' % dst)

        # Pull the neighbor MAC address from the IP neighbor table
        try:
            nbrline = subprocess.check_output(['ip', 'neighbor', 'show',
                                               dst]).strip()
            if not nbrline:
                raise NetQException('No ARP entry found for IP %s' % dst)
        except:
            raise NetQException('Unable to find ARP entry for IP %s' % dst)
        words = nbrline.split()
        if words[-1] == 'FAILED' or words[-1] == 'INCOMPLETE':
            raise NetQException('No ARP entry found for IP %s' % dst)
        mac_address = words[4]

        # Get the src VXLAN device
        for smacaddr in MacFdb.query.filter(timestamp=self.__start_time,
                                            mac_address=mac_address):
            if not smacaddr.dst:
                continue
            """:type : VxlanLink"""
            srcvxlandev = next(
                VxlanLink.query.filter(timestamp=self.__start_time,
                                       ifname=smacaddr.nexthop,
                                       hostname=smacaddr.hostname),
                None)
            if srcvxlandev is not None:
                break
        else:
            raise NetQException(
                'Couldn\'t locate VxlanDev obj for mac address %s' %
                mac_address
            )

        # Get the dst VXLAN device
        for dstvxlandev in VxlanLink.query.filter(timestamp=self.__start_time,
                                                  vni=srcvxlandev.vni):
            if dstvxlandev.localip == smacaddr.dst:
                break
        else:
            raise NetQException(
                'Couldn\'t locate dst VxlanDev obj for vni %s' %
                srcvxlandev.vni
            )

        # Finally, execute a traceroute
        cmd = '/usr/sbin/traceroute -n %s' % dstvxlandev.localip
        tracecmd2 = subprocess.Popen(cmd.split(),
                                     stdout=subprocess.PIPE,
                                     stderr=subprocess.STDOUT)
        idx = 0
        out = []
        for line in iter(tracecmd2.stdout.readline, b''):
            if first_line:
                out.append(first_line.rstrip())
                first_line = ''
                out.append(
                    '---------------------- Underlay -------------------------'
                    '--------'
                )
            else:
                match = re.findall(self.__TRACEROUTE_REGEX, line)
                if match:
                    idx = match[0]
                out.append(line.rstrip())
        if last_line:
            oidx = re.findall(self.__TRACEROUTE_REGEX, last_line)[0]
            out.append(
                '-------------------------------------------------------------'
                '----'
            )
            out.append(last_line.replace(oidx, str(int(idx) + 1), 1).rstrip())
        return '\n'.join(out)

    @staticmethod
    def update_config(server_address):
        try:
            redisdb.ping_server(server_address)
        except Exception:  # pylint: disable=broad-except
            raise RuntimeError('Could not connect to server %s' %
                               server_address)
        conf = config.Config(DefaultPaths.CONF_FILE)
        try:
            conf.set_field(config.Config.CommonConfig.section,
                           config.Config.CommonConfig.server.name,
                           server_address)
        except Exception as ex:
            raise RuntimeError('Failed to update config file %s. Error %s' %
                               (DefaultPaths.CONF_FILE, ex))
        return 'Success!'

    ##########################
    # TAB Completion functions
    ##########################
    @staticmethod
    def view_complete(args, ended_with_space, last_argv):
        """Return possible completions for current pos.
        """
        server_addr = NetQ.__get_config_server_addr()
        if server_addr is None:
            return
        if server_addr is not None and 'http' in server_addr:
            restapi.set_connection_settings(server_addr)
        else:
            redisdb.set_connection_settings(server_addr)
        hostname = args.get('<hostname>')
        if not hostname:
            return []

        args = args.get('<wildcard-command>') or ''
        args = args.split()
        if not ended_with_space:
            args = args[:-1]
        if args:
            pos = len(args)
        else:
            args = []
            pos = 0
        key = ''
        if pos >= 1:
            key = '-'.join(str(arg) for arg in args)
        key += '*'
        choices = []
        for cmd in Command.query.filter(hostname=hostname, key=key):
            cmdlist = cmd.key.split('-')
            if len(cmdlist) == pos:
                nextarg = cmd.key.split('-')[pos - 1]
                if ended_with_space or nextarg.startswith(last_argv):
                    if nextarg != last_argv:
                        # Don;t add the same option back. This is to handle
                        # cases such as 'mstp bridge [json]' and the cursor is
                        # at bridge. Since this is a valid option, we'll end up
                        # displaying bridge again.
                        choices.append((nextarg, nextarg))
            else:
                if pos == 0:
                    nextarg = cmd.key.split('-')[pos]
                else:
                    nextarg = cmd.key.split('-')[pos]
                if ended_with_space or nextarg.startswith(last_argv):
                    choices.append((nextarg, nextarg))

        if not choices:
            choices.append(('<wildcard-command>', 'Remote Command'))
        return sorted(choices)

    @staticmethod
    def view_match(argv_word, use_fuzzy, tab_completing=False):
        """
        Matching against a command is impossible when we only have one word
        """
        server_addr = NetQ.__get_config_server_addr()
        if server_addr is None:
            return
        if server_addr is not None and 'http' in server_addr:
            restapi.set_connection_settings(server_addr)
        else:
            redisdb.set_connection_settings(server_addr)

        if sys.argv[1] == 'server':
            args = sys.argv[4:]
        else:
            args = sys.argv[2:]

        # The first arg is the hostname
        hostname = args[0]
        # args = args[1:-1]
        args = args[1:]

        # dwalton
        if args:
            pos = len(args)
        else:
            pos = 0
        key = ''

        if pos >= 1:
            key = '-'.join(map(str, args))
            key += '*'
        else:
            key += '%s*' % argv_word

        if Command.query.filter(hostname=hostname, key=key):
            return True

    @staticmethod
    def hostname_complete(args, ended_with_space, last_argv):
        """
        Completion for hostname
        """
        out = []
        server_addr = NetQ.__get_config_server_addr()
        if server_addr is None:
            return []
        if server_addr is not None and 'http' in server_addr:
            restapi.set_connection_settings(server_addr)
        else:
            redisdb.set_connection_settings(server_addr)
        for node in Node.query.all():
            if ended_with_space or node.hostname.startswith(last_argv):
                out.append((node.hostname, node.hostname))

        if not out:
            out.append(('<hostname>', 'Remote hostname'))

        return out

    @staticmethod
    def hostname_match(argv_word, use_fuzzy, tab_completing=False):
        """
        <as-path-access-list>
        """
        server_addr = NetQ.__get_config_server_addr()
        if server_addr is None:
            return True
        if server_addr is not None and 'http' in server_addr:
            restapi.set_connection_settings(server_addr)
        else:
            redisdb.set_connection_settings(server_addr)

        node_lst = sorted(node.hostname for node in Node.query.all())
        if use_fuzzy:
            if ' ' not in argv_word:
                return True
        else:
            for node in node_lst:
                if tab_completing and node.startswith(argv_word):
                    return True
                if argv_word in node:
                    return True
        return False

    @staticmethod
    def interface_complete(args, ended_with_space, last_argv):
        """
        Completion for remote interface name
        """
        out = []
        server_addr = NetQ.__get_config_server_addr()
        if server_addr is None:
            return []
        if server_addr is not None and 'http' in server_addr:
            restapi.set_connection_settings(server_addr)
        else:
            redisdb.set_connection_settings(server_addr)

        hostname = args.get('<hostname>')
        if not hostname:
            return []

        plinks = Link.query.filter(hostname=hostname)
        if not args.get('lldp'):
            vlinks = VirtualDevLink.query.filter(hostname=hostname)
            vxlinks = VxlanLink.query.filter(hostname=hostname)
            links = chain(plinks, vlinks, vxlinks)
        else:
            links = plinks
            
        for link in links:
            if args.get('lldp'):
                if link.ifname == 'lo':
                    continue

                if link.kind in ['vlan', 'vxlan', 'macvlan']:
                    continue
            
            if ended_with_space or link.ifname.startswith(last_argv):
                out.append((link.ifname, link.ifname))

        if not out:
            out.append(('<remote-interface>', 'Interface name on %s' %
                        hostname))
        return out

    @staticmethod
    def interface_match(argv_word, use_fuzzy, tab_completing=False):
        """
        Match for remote interface name
        """
        keywords = ['history', 'origin', 'json', 'vlan']
        if argv_word in keywords:
            return False

        try:
            if ipaddr.IPAddress(argv_word):
                return False
        except ValueError:
            pass

        try:
            if ipaddr.IPNetwork(argv_word):
                return False
        except ValueError:
            if (re.match(r'(?:[0-9A-Fa-f]{2}[:-]){5}[0-9A-Fa-f]{2}',
                         argv_word) or
                    re.match(r'[0-9A-Fa-f]{4}\.[0-9A-Fa-f]{4}\.[0-9A-Fa-f]{4}',
                             argv_word)):
                return False

        server_addr = NetQ.__get_config_server_addr()
        if server_addr is None:
            return []
        if server_addr is not None and 'http' in server_addr:
            restapi.set_connection_settings(server_addr)
        else:
            redisdb.set_connection_settings(server_addr)

        hostname = get_hostname()
        if not hostname:
            return []

        plinks = Link.query.filter(hostname=hostname)
        if not 'lldp' in sys.argv:
            vlinks = VirtualDevLink.query.filter(hostname=hostname)
            vxlinks = VxlanLink.query.filter(hostname=hostname)
            links = chain(plinks, vlinks, vxlinks)
        else:
            links = plinks
            
        for link in links:
            return True

        return False

    @property
    def node_lst(self):
        if self.__node_lst is None:
            self.__node_lst = sorted(Node.query.all())
        return self.__node_lst
