# vim: tabstop=4 shiftwidth=4 softtabstop=4
# Copyright 2016 Cumulus Networks, Inc. All rights reserved.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc.
# 51 Franklin Street, Fifth Floor
# Boston, MA  02110-1301, USA.
from __future__ import absolute_import

import json

from netq.orm.memdb.models import (
    BgpSession,
    Lldp,
    MemNode
)


def build_bgp_summary(hostname, timestamp, cmd_output, neigh_output=None):
    """Build summary of BGP node info, such as ASN & RID
    """
    json_out = json.loads(cmd_output)
    bgp_inactive_peer_count = 0

    # Save node specific info
    try:
        node = next(MemNode.query.filter(hostname=hostname))
    except StopIteration:
        node = MemNode()
        node.hostname = hostname
        node.timestamp = timestamp

    if isinstance(json_out['routerId'], unicode):
        node.bgp_router_id = (
            json_out['routerId'].encode('ascii', 'ignore')
        )
    else:
        node.bgp_router_id = json_out['routerId']
    node.bgp_asn = json_out['as']
    node.bgp_peer_count = json_out['peerCount']
    node.bgp_rib_entries = json_out['ribCount']
    node.bgp_rib_mem = json_out['ribMemory']

    # Now also save BGP session state
    for peer in json_out['peers']:
        bgp = BgpSession()
        bgp.hostname = hostname
        bgp.timestamp = timestamp
        if isinstance(peer, unicode):
            bgp.peer_name = peer.encode('ascii', 'ignore')
        else:
            bgp.peer_name = peer
        if isinstance(json_out['peers'][peer]['state'], unicode):
            bgp.peer_state = (
                json_out['peers'][peer]['state'].encode('ascii',
                                                        'ignore')
            )
        else:
            bgp.peer_state = json_out['peers'][peer]['state']
        if neigh_output is not None and bgp.peer_state != 'Established':
            match = False
            bgp_inactive_peer_count += 1
            for line in neigh_output.splitlines():
                line = line.strip()
                bgp.reason = ''
                if line.startswith('BGP neighbor on %s:' % bgp.peer_name):
                    match = True
                elif match:
                    if line.startswith('Connections established:'):
                        bgp.reason += (line + ' ')
                    if line.startswith('Last reset'):
                        bgp.reason += (line + ' ')
                        break
                elif line.startswith('BGP neighbor on'):
                    match = False
        else:
            bgp.reason = ''
        bgp.peer_asn = json_out['peers'][peer]['remoteAs']
        bgp.pfx_rcv_count = json_out['peers'][peer]['prefixReceivedCount']
        bgp.save()

    node.bgp_active_peer_count = node.bgp_peer_count - bgp_inactive_peer_count
    node.save()


def build_clag_map(hostname, timestamp, cmd_output):
    """Build clag info for node
    """
    # Save node specific info
    json_out = json.loads(cmd_output)

    try:
        node = next(MemNode.query.filter(hostname=hostname))
    except StopIteration:
        node = MemNode()
        node.timestamp = timestamp
        node.hostname = hostname

    node.clag_state = True
    node.clag_peer_state = json_out['status']['peerAlive']
    if node.clag_peer_state != '1':
        node.clag_state = False
    node.clag_peer_if = json_out['status']['peerIf'].encode('ascii', 'ignore')
    node.clag_peer_role = json_out['status']['peerRole'].encode('ascii',
                                                                'ignore')
    node.clag_role = json_out['status']['ourRole'].encode('ascii', 'ignore')
    node.clag_sysmac = json_out['status']['sysMac'].encode('ascii', 'ignore')
    node.clag_backup_ip = json_out['status']['backupIp'].encode('ascii',
                                                                'ignore')
    clag_conflicted_bonds = []
    clag_singly_attached_bonds = []
    clag_proto_down_bonds = []
    for iface in json_out['clagIntfs']:
        if json_out['clagIntfs'][iface]['status'] != 'dual':
            clag_singly_attached_bonds.append(iface)
        if 'conflicts' in json_out['clagIntfs'][iface]:
            clag_conflicted_bonds.append(
                [iface,
                 json_out['clagIntfs'][iface]['conflicts'][0].encode('ascii',
                                                                  'ignore')])
            node.clag_state = False
        if 'protoDown' in json_out['clagIntfs'][iface]:
            string = (
                json_out['clagIntfs'][iface]['protoDown'][0].encode('ascii',
                                                                    'ignore')
            )
            clag_proto_down_bonds.append([iface, string])
            node.clag_state = False

    #This weird way is due to some ORM stuff. FIXME Sameer
    node.clag_conflicted_bonds = clag_conflicted_bonds
    node.clag_singly_attached_bonds = clag_singly_attached_bonds
    node.clag_proto_down_bonds = clag_proto_down_bonds

    node.save()


def build_lldp_summary(hostname, timestamp, cmd_output):
    json_op = json.loads(cmd_output)
    for key, values in json_op.iteritems():
        for value in values:
            try:
                for lldp in value['interface']:
                    lldpobj = Lldp()
                    lldpobj.hostname = hostname
                    lldpobj.timestamp = timestamp

                    try:
                        lldpobj.local_intf = str(lldp['name'])
                        lldpobj.lldp_peer = str(
                            lldp['chassis'][0]['name'][0]['value']
                        )
                        if '.' in lldpobj.lldp_peer:
                            lldpobj.lldp_peer = (
                                lldpobj.lldp_peer.split('.')[0]
                            )
                        lldpobj.lldp_intf = str(
                            lldp['port'][0]['id'][0]['value']
                        )

                        for cap in lldp['chassis'][0]['capability']:
                            if cap['type'] == 'Bridge':
                                if cap['enabled']:
                                    lldpobj.lldp_peer_bridge = True
                                else:
                                    lldpobj.lldp_peer_bridge = False

                            elif cap['type'] == 'Router':
                                if cap['enabled']:
                                    lldpobj.lldp_peer_router = True
                                else:
                                    lldpobj.lldp_peer_router = False

                        lldpobj.lldp_peer_os = str(
                            lldp['chassis'][0]['descr'][0]['value']
                        )
                    except Exception:
                        continue
                    else:
                        lldpobj.save()
            except Exception:
                continue
