# vim: tabstop=4 shiftwidth=4 softtabstop=4
# Copyright 2016 Cumulus Networks, Inc. All rights reserved.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc.
# 51 Franklin Street, Fifth Floor
# Boston, MA  02110-1301, USA.
"""ROM models
"""
import time

from netq.orm import objects, HOSTNAME
from netq.orm.redisdb import Connection, scripts
from netq.orm.redisdb.query import RedisQuery


class _RedisModel(objects.Model):
    query_cls = RedisQuery
    active = objects.Boolean(json_key='active')
    hostname = objects.String(json_key='hostname')
    timestamp = objects.Float(json_key='timestamp')

    def __init__(self):
        self.active = True
        self.hostname = HOSTNAME
        self.timestamp = time.time()
        super(_RedisModel, self).__init__()

    @staticmethod
    def _get_connection_object(commit):
        if commit:
            return Connection.REDIS
        else:
            return Connection.PIPE

    def delete(self, purge=False):
        if purge:
            score = self.__class__.obj_to_score(self.__dict__)
            Connection.PIPE.zremrangebyscore(
                self.__class__.obj_to_key(self.__dict__),
                min=score,
                max=score
            )
        else:
            obj = self.__class__()
            for field in vars(self.__class__).itervalues():
                if (isinstance(field, objects.Field) and
                        field.name != self.__class__.timestamp.name):
                    setattr(obj, field.name, getattr(self, field.name))
            obj.active = False
            obj.deleted = True
            obj.save()

    @staticmethod
    def save_entry(connection, key, score, val, update=False):
        if update:
            scripts.update_entry_script(keys=[key], args=[score, val],
                                        client=connection)
        else:
            connection.zadd(key, score, val)


class Address(_RedisModel):
    ifname = objects.String(json_key='ifName')
    prefix = objects.String(json_key='prefix')
    mask = objects.Integer(json_key='mask')
    is_ipv6 = objects.Boolean(json_key='isIPv6')
    deleted = objects.Boolean(json_key='deleted')

    @classmethod
    def key_fmt(cls, key_type=None):
        return (cls.__name__, cls.hostname, cls.ifname, cls.prefix,
                cls.mask, cls.is_ipv6)

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def val_fmt(cls):
        return cls.timestamp, cls.active, cls.deleted

    def save(self, commit=False):
        connection = self._get_connection_object(commit)
        key = Address.obj_to_key(self.__dict__)
        score = Address.obj_to_score(self.__dict__)
        val = Address.obj_to_val(self.__dict__)
        super(Address, self).save_entry(connection, key, score, val)


class Command(_RedisModel):
    key = objects.String(json_key='key')
    output = objects.String(json_key='output')

    @classmethod
    def key_fmt(cls, key_type=None):
        return cls.__name__, cls.hostname, cls.key

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def val_fmt(cls):
        return cls.output,

    def save(self, commit=False):
        connection = self._get_connection_object(commit)
        connection.zadd(Command.obj_to_key(self.__dict__),
                        Command.obj_to_score(self.__dict__),
                        Command.obj_to_val(self.__dict__))


class Counter(_RedisModel):
    ifname = objects.String(json_key='ifName')
    rx_bytes = objects.Integer(json_key='rxBytes')
    rx_drop_bytes = objects.Integer(json_key='rxDropBytes')
    tx_bytes = objects.Integer(json_key='txBytes')
    tx_drop_bytes = objects.Integer(json_key='txDropBytes')

    TX_BYTES = 0
    TX_DROP_BYTES = 1
    RX_BYTES = 2
    RX_DROP_BYTES = 3

    @classmethod
    def key_fmt(cls, key_type=None):
        if key_type == cls.TX_BYTES:
            return cls.__name__, cls.tx_bytes.name
        elif key_type == cls.TX_DROP_BYTES:
            return cls.__name__, cls.tx_drop_bytes.name
        elif key_type == cls.RX_BYTES:
            return cls.__name__, cls.rx_bytes.name
        elif key_type == cls.RX_DROP_BYTES:
            return cls.__name__, cls.rx_drop_bytes.name
        else:
            raise RuntimeError('Invalid key type %s' % key_type)

    @classmethod
    def score_fmt(cls, key_type=None):
        key_type = key_type or cls.TX_BYTES
        if key_type == cls.TX_BYTES:
            return cls.tx_bytes
        elif key_type == cls.TX_DROP_BYTES:
            return cls.tx_drop_bytes
        elif key_type == cls.RX_BYTES:
            return cls.rx_bytes
        elif key_type == cls.RX_DROP_BYTES:
            return cls.rx_drop_bytes

    @classmethod
    def val_fmt(cls):
        return cls.__name__, cls.hostname, cls.ifname

    def save(self, commit=False):
        connection = self._get_connection_object(commit)
        intf_key = Counter.obj_to_val(self.__dict__)
        connection.zadd(Counter.obj_to_key(self.__dict__,
                                           key_type=Counter.RX_BYTES),
                        Counter.obj_to_score(self.__dict__,
                                             key_type=
                                             Counter.RX_BYTES),
                        intf_key)
        connection.zadd(Counter.obj_to_key(self.__dict__,
                                           key_type=
                                           Counter.RX_DROP_BYTES),
                        Counter.obj_to_score(self.__dict__,
                                             key_type=
                                             Counter.RX_DROP_BYTES),
                        intf_key)
        connection.zadd(Counter.obj_to_key(self.__dict__,
                                           key_type=Counter.TX_BYTES),
                        Counter.obj_to_score(self.__dict__,
                                             key_type=
                                             Counter.TX_BYTES),
                        intf_key)
        connection.zadd(Counter.obj_to_key(self.__dict__,
                                           key_type=
                                           Counter.TX_DROP_BYTES),
                        Counter.obj_to_score(self.__dict__,
                                             key_type=
                                             Counter.TX_DROP_BYTES),
                        intf_key)


class Link(_RedisModel):
    ifname = objects.String(json_key='ifName')
    ifindex = objects.Integer(json_key='ifIndex')
    mac_address = objects.MacAddress(json_key='macAddress')
    admin_state = objects.String(json_key='adminState')
    oper_state = objects.String(json_key='operState')
    managed = objects.Boolean(json_key='managed')
    kind = objects.String(json_key='kind')
    deleted = objects.Boolean(json_key='deleted')
    mtu = objects.Integer(json_key='mtu')
    is_slave = objects.Boolean(json_key='isSlave')

    @classmethod
    def key_fmt(cls, key_type=None):
        return cls.__name__, cls.hostname, cls.ifname, cls.kind, cls.mac_address

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def val_fmt(cls):
        return (cls.admin_state, cls.oper_state, cls.managed, cls.mtu,
                cls.ifindex, cls.is_slave, cls.timestamp, cls.active,
                cls.deleted)

    def save(self, commit=False):
        connection = self._get_connection_object(commit)
        key = Link.obj_to_key(self.__dict__)
        score = Link.obj_to_score(self.__dict__)
        val = Link.obj_to_val(self.__dict__)
        super(Link, self).save_entry(connection, key, score, val)


class VirtualDevLink(_RedisModel):
    """Different from the regular link because key characteristics change
    """
    ifname = objects.String(json_key='ifName')
    ifindex = objects.Integer(json_key='ifIndex')
    mac_address = objects.MacAddress(json_key='macAddress')
    admin_state = objects.String(json_key='adminState')
    oper_state = objects.String(json_key='operState')
    managed = objects.Boolean(json_key='managed')
    kind = objects.String(json_key='kind')
    deleted = objects.Boolean(json_key='deleted')
    mtu = objects.Integer(json_key='mtu')
    slaves = objects.Json(json_key='slaves')
    is_slave = objects.Boolean(json_key='isSlave')

    @classmethod
    def key_fmt(cls, key_type=None):
        return cls.__name__, cls.hostname, cls.ifname

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def val_fmt(cls):
        return (cls.admin_state, cls.oper_state, cls.managed, cls.mtu,
                cls.ifindex, cls.kind, cls.timestamp, cls.mac_address, cls.slaves,
                cls.active, cls.deleted)

    def save(self, commit=False):
        connection = self._get_connection_object(commit)
        key = VirtualDevLink.obj_to_key(self.__dict__)
        score = VirtualDevLink.obj_to_score(self.__dict__)
        val = VirtualDevLink.obj_to_val(self.__dict__)
        super(VirtualDevLink, self).save_entry(connection, key, score, val)


class MacFdb(_RedisModel):
    mac_address = objects.MacAddress(json_key='macAddress')
    origin = objects.Boolean(json_key='origin')
    nexthop = objects.String(json_key='nextHop')
    vlan = objects.Integer(json_key='vlan')
    dst = objects.String(json_key='dst')
    port = objects.String(json_key='port')
    deleted = objects.Boolean(json_key='deleted')

    @classmethod
    def key_fmt(cls, key_type=None):
        return (cls.__name__, cls.hostname, cls.mac_address, cls.vlan,
                cls.nexthop)

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def val_fmt(cls):
        return (cls.nexthop, cls.origin, cls.dst, cls.port, cls.timestamp,
                cls.active, cls.deleted)

    def save(self, commit=False):
        connection = self._get_connection_object(commit)
        key = MacFdb.obj_to_key(self.__dict__)
        score = MacFdb.obj_to_score(self.__dict__)
        val = MacFdb.obj_to_val(self.__dict__)
        super(MacFdb, self).save_entry(connection, key, score, val)


class Neighbor(_RedisModel):
    ifname = objects.String(json_key='ifName')
    ifindex = objects.Integer(json_key='ifIndex')
    ip_address = objects.String(json_key='ipAddress')
    mac_address = objects.MacAddress(json_key='macAddress')
    is_ipv6 = objects.Boolean(json_key='isIPv6')
    deleted = objects.Boolean(json_key='deleted')

    @classmethod
    def key_fmt(cls, key_type=None):
        return (cls.__name__, cls.hostname, cls.ifname,
                cls.ip_address, cls.mac_address, cls.is_ipv6)

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def val_fmt(cls):
        return cls.ifindex, cls.timestamp, cls.active, cls.deleted

    def save(self, commit=False):
        connection = self._get_connection_object(commit)
        key = Neighbor.obj_to_key(self.__dict__)
        score = Neighbor.obj_to_score(self.__dict__)
        val = Neighbor.obj_to_val(self.__dict__)
        super(Neighbor, self).save_entry(connection, key, score, val)


class Node(_RedisModel):
    lastboot = objects.Float(json_key='lastBoot')

    def __str__(self):
        return self.hostname

    @classmethod
    def key_fmt(cls, key_type=None):
        return cls.__name__, cls.hostname

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def val_fmt(cls):
        return cls.lastboot,

    def save(self, commit=False):
        connection = self._get_connection_object(commit)
        key = Node.obj_to_key(self.__dict__)
        score = Node.obj_to_score(self.__dict__)
        val = Node.obj_to_val(self.__dict__)
        super(Node, self).save_entry(connection, key, score, val)


class Route(_RedisModel):
    ip_address = objects.String(json_key='ipAddress')
    protocol = objects.String(json_key='protocol')
    origin = objects.Boolean(json_key='origin')
    nexthops = objects.Json(json_key='nextHops')
    src = objects.String(json_key='src')
    is_ipv6 = objects.Boolean(json_key='isIPv6')
    deleted = objects.Boolean(json_key='deleted')
    table = objects.Integer(json_key='table')
    family = objects.Integer(json_key='family')

    @classmethod
    def key_fmt(cls, key_type=None):
        return (cls.__name__, cls.hostname, cls.ip_address, cls.is_ipv6,
                cls.table, cls.family, cls.origin)

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def val_fmt(cls):
        return (cls.nexthops, cls.origin, cls.protocol, cls.src, cls.timestamp,
                cls.active, cls.deleted, cls.table, cls.family)

    def save(self, commit=False):
        connection = self._get_connection_object(commit)
        key = Route.obj_to_key(self.__dict__)
        score = Route.obj_to_score(self.__dict__)
        val = Route.obj_to_val(self.__dict__)
        super(Route, self).save_entry(connection, key, score, val)


class VxlanLink(_RedisModel):
    brifindex = objects.Integer(json_key='bridgeIfIndex')
    dstport = objects.Integer(json_key='dstPort')
    ifname = objects.String(json_key='ifName')
    ifindex = objects.Integer(json_key='ifIndex')
    localip = objects.String(json_key='localIp')
    mac_address = objects.MacAddress(json_key='macAddress')
    managed = objects.Boolean(json_key='managed')
    admin_state = objects.String(json_key='adminState')
    oper_state = objects.String(json_key='operState')
    remoteip = objects.String(json_key='remoteIp')
    vni = objects.Integer(json_key='vni')
    deleted = objects.Boolean(json_key='deleted')
    mtu = objects.Integer(json_key='mtu')
    kind = objects.String(json_key='kind')
    is_slave = objects.Boolean(json_key='isSlave')

    @classmethod
    def key_fmt(cls, key_type=None):
        return cls.__name__, cls.hostname, cls.ifname, cls.vni

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def val_fmt(cls):
        return (cls.admin_state, cls.oper_state, cls.managed, cls.dstport,
                cls.localip, cls.remoteip, cls.brifindex, cls.mac_address,
                cls.timestamp, cls.kind, cls.is_slave, cls.active,
                cls.deleted, cls.mtu)

    def save(self, commit=False):
        connection = self._get_connection_object(commit)
        key = VxlanLink.obj_to_key(self.__dict__)
        score = VxlanLink.obj_to_score(self.__dict__)
        val = VxlanLink.obj_to_val(self.__dict__)
        super(VxlanLink, self).save_entry(connection, key, score, val)
