-- vim: tabstop=4 shiftwidth=4 softtabstop=4
-- Copyright 2016 Cumulus Networks, Inc. All rights reserved.
--
-- This program is free software; you can redistribute it and/or
-- modify it under the terms of the GNU General Public License
-- as published by the Free Software Foundation; either version 2
-- of the License, or (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software
-- Foundation, Inc.
-- 51 Franklin Street, Fifth Floor
-- Boston, MA  02110-1301, USA
function keys_script()
    local match = KEYS[1]
    local tmp, keys
    local cur = 0
    local result = {}
    repeat
        tmp = redis.call("SCAN", cur, "MATCH", match)
        cur = tonumber(tmp[1])
        keys = tmp[2]
        if keys ~= nil then
            for idx = 1, #keys do
                result[#result + 1] = keys[idx]
            end
        end
    until cur == 0
    return result
end


function delregex_script()
    local new_score = ARGV[1]
    local purge = ARGV[2]
    local tmp, score, val, val_json
    for idx = 1, #KEYS do
        tmp = redis.call("ZRANGE", KEYS[idx], -1, -1, "WITHSCORES")
        val = tmp[1]
        score = tonumber(tmp[2])
        if val ~= nil then
            if purge == 1 then
                redis.call("DEL", KEYS[idx])
            else
                val_json = cjson.decode(val)
                if val_json["active"] ~= "0" then
                    val_json["active"] = "0"
                    val_json["timestamp"] = new_score
                    redis.call("ZADD", KEYS[idx], new_score, cjson.encode(val_json))
                end
            end
        end
    end
end


function gc_script()
    local freed = 0
    local count = tonumber(ARGV[1])
    for index = 1, #KEYS do
        local currentcount = tonumber(redis.call("ZCARD", KEYS[index]))
        if currentcount >= count then
            freed = freed + currentcount - count
            redis.call("ZREMRANGEBYRANK", KEYS[index], 0, currentcount - count)
        end
    end
    return freed
end

function lpm_script()
    local hostnames = KEYS
    local match = ARGV[1]
    local input_prefix = ARGV[2]
    local start_time = ARGV[3]
    local is_ipv6 = tonumber(ARGV[4])
    local seperator = ARGV[5]

    -- Expand the prefix if it is IPv6
    if is_ipv6 == 1 then
        input_prefix = utils.expand_ipv6(input_prefix)
    end

    -- Regex construction
    local ipv4seg = "%d%d?%d?"
    local ipv6seg = "[0-9a-fA-F]?[0-9a-fA-F]?[0-9a-fA-F]?[0-9a-fA-F]?:?:?"
    local remask = "%d+"
    -- We assume the hostname comes before the IP address
    local reipv4addr = string.format(seperator.."(.+)"..seperator.."(%s)/(%s)",
                                     string.rep(ipv4seg.."%.", 3)..ipv4seg,
                                     remask)
    local reipv6addr = string.format(seperator.."(.+)"..seperator.."(%s)/(%s)",
                                     string.rep(ipv6seg, 8),
                                     remask)

    -- Variable initialization
    local src_prefix_len, src_prefix_inc
    if is_ipv6 == 1 then
        src_prefix_len = 128
        src_prefix_inc = 16
    else
        src_prefix_len = 32
        src_prefix_inc = 8
    end

    local tmp, longest_key, longest_val, longest_score
    local results = {}
    local found = {}
    local key = string.gsub(match, "(%/%d?%d?)$", "")
    local prefix_len = src_prefix_len
    local prefix_str = input_prefix
    while prefix_str ~= "*" do
        prefix_len = prefix_len - src_prefix_inc
        prefix_str = utils.sub_prefix(prefix_str, is_ipv6)
        local new_key = string.gsub(key, seperator..input_prefix, seperator..prefix_str)
        local cur = 0
        repeat
            tmp = redis.call("SCAN", cur, "MATCH", new_key)
            cur = tonumber(tmp[1])
            local keys = tmp[2]
            local hostname, prefix, mask
            if keys then
                for idx = 1, #keys do
                    if is_ipv6 == 1 then
                        hostname, prefix, mask = string.match(keys[idx], reipv6addr)
                    else
                        hostname, prefix, mask = string.match(keys[idx], reipv4addr)
                    end
                    mask = tonumber(mask)
                    if hostname~= nil and prefix ~= nil and mask ~= nil and
                       mask >= prefix_len and mask <= prefix_len + src_prefix_inc and
                       utils.is_match(prefix, input_prefix, mask, is_ipv6) and
                       (found[hostname] == nil or mask > found[hostname][1]) then
                        longest_key = keys[idx]
                        found[hostname] = {mask, longest_key}
                    end
                end
            end
        until cur == 0
        tmp = 0
        for idx = 1, #hostnames do
            if found[hostnames[idx]] then
                tmp = tmp + 1
            end
        end
        if tmp == #hostnames then
            prefix_str = "*"
        end
    end
    for k, v in pairs(found) do
        for idx = 1, #hostnames do
            if k == hostnames[idx] then
                longest_key = v[2]
                tmp = redis.call("ZREVRANGEBYSCORE", longest_key, start_time, 0, "WITHSCORES")
                if tmp[1] then
                    longest_val = tmp[1]
                    longest_score = tmp[2]
                    results[#results + 1] = {longest_key, longest_val, longest_score }
                end
                break
            end
        end
    end
    return results
end

function scanregex_script()
    local match = KEYS[1]
    local nmatch = KEYS[2]
    local cur = tonumber(KEYS[3])
    local count = KEYS[4]
    local rangeapi = ARGV[1]
    local rangestart = ARGV[2]
    local rangeend = ARGV[3]
    local first = tonumber(ARGV[4])
    local rep = {}
    local tmp, keys
    repeat
        tmp = redis.call("SCAN", cur, "MATCH", match)
        cur = tonumber(tmp[1])
        keys = tmp[2]
        if keys then
            for idx = 1, #keys do
                local key_type = redis.call("TYPE", keys[idx])
                if type(key_type) == "table" then
                    -- hacky conversion from table to string
                    for _, tmp in pairs(key_type) do
                        key_type = tmp
                    end
                end
                if nmatch == ".*" or not keys[idx]:find(nmatch) then
                    if key_type == "set" or key_type == "string" then
                        rep[#rep + 1] = {keys[idx], redis.call("GET", keys[idx]), 0}
                    elseif key_type == "hash" then
                        tmp = redis.call("HGETALL", keys[idx])
                        for index = 1, #tmp do
                            rep[#rep + 1] = {keys[idx], tmp[index], 0}
                        end
                    elseif key_type == "zset" then
                        tmp = redis.call(rangeapi, keys[idx], rangestart, rangeend, "WITHSCORES")
                        for index=1, #tmp, 2 do
                            rep[#rep + 1] = {keys[idx], tmp[index], tmp[index + 1]}
                            if first == 1 then
                                break
                            end
                        end
                    end
                end
            end
        end
    until cur == 0 or (count and #rep == tonumber(count))
    return {cur, rep}
end

function update_entry_script()
    local key = KEYS[1]
    local new_score = ARGV[1]
    local new_val = ARGV[2]
    local new_val_json = cjson.decode(new_val)
    local score, tmp, val, val_json
    local cmp = 0
    tmp = redis.call("ZRANGE", key, -1, -1, "WITHSCORES")
    val = tmp[1]
    if val ~= nil then
        val_json = cjson.decode(val)
        -- Compare the old and new values (everything besides active)
        for k1, v1 in pairs(val_json) do
            if k1 ~= "active" and k1 ~= "timestamp" then
                cmp = 0
                for k2, v2 in pairs(new_val_json) do
                    if k1 == k2 and v1 == v2 then
                        cmp = 1
                        break
                    end
                end
                if cmp == 0 then
                    break
                end
            end
        end
    end

    if cmp == 0 then
        redis.call("ZADD", key, new_score, new_val)
    elseif new_val_json["active"] == "1" and val_json["active"] == "0" then
        redis.call("ZREMRANGEBYRANK", key, -1, -1)
    end
end
